#include <iostream>
#include <iomanip>
#include <fstream>
#include <string>
#include <cassert>
#include <cmath>

#include "vector.h"
#include "pixel.h"
#include "camera.h"
#include "light.h"
#include "material.h"
#include "object.h"
#include "list.h"
#include "plane.h"
#include "model.h"

// prototypes
void ray_trace(model_t&,vec_t&,vec_t&,rgb_t<double>&,double,object_t*);
int  main(int argc, char *argv[]);

int main(int argc, char *argv[])
{
	model_t		model;		// model (the world)
	std::ifstream	model_ifs;	// input file stream

  if(argc != 2) {
    std::cerr << "Usage " << argv[0] << " <model_file>" << std::endl;
    return 1;
  }

  // open file, read in model
  model_ifs.open(argv[1],std::ifstream::in);
  model_ifs >> model;
  model_ifs.close();

  // debugging
  std::cerr << model;

	int		i,x,y;
	int		w=model.getpixel_w(), h=model.getpixel_h();
	double		wx,wy,wz=0.0;
	double		ww=model.getworld_w(), wh=model.getworld_h();
	vec_t		pos,pix,dir;
	rgb_t<double>	color;
	rgb_t<uchar>	icolor;
	unsigned char	*imgloc,*img=NULL;

  img = new unsigned char[w*h];

  std::cout << "P6 " << w << " " << h << " " << 255 << std::endl;
  for(y=h-1;y>=0;y--) {
    for(x=0;x<w;x++) {
      wx = (double)x/(double)(w-1) * ww;
      wy = (double)y/(double)(h-1) * wh;

      // copy vector v1 to vector v2
      pos = model.getviewpoint();

      // set pixel position vector (in world coordinates)
      pix[0] = wx;  pix[1] = wy;  pix[2] = wz;

      // compute the vector difference  v3 = v2 - v1
      dir = pix - pos;

      // our ray is now {pos, dir} (in world coordinates), normalize dir
      dir = dir.norm();

      // zero out color
      color[0] = 0.0;  color[1] = 0.0;  color[2] = 0.0;

      // trace ray
      ray_trace(model,pos,dir,color,0.0,NULL);

      // scale pixel by maxval
      color = 255.0 * color;

      // where are we in the image (using old i*c + j)
    //imgloc = img + y*w + x;
    //for(i=0;i<3;i++) imgloc[i] = color[i];
    //std::cout.write((char *)imgloc,3);

      // instead of using image, just write pixel to stdout
      for(i=0;i<3;i++) icolor[i] = static_cast<uchar>(color[i]);
      for(i=0;i<3;i++) std::cout.put(icolor[i]);
    }
  }

  return 0;
}

#define MAX_DIST 100

void ray_trace(model_t&		model,
               vec_t&		pos,
               vec_t&		dir,
               rgb_t<double>&	color,
               double		raydist,
               object_t		*last_hit)
{
	double				dist=0.0;
	object_t			*obj=NULL;
	rgb_t<double>			ambient, diffuse, specular;
	material_t			*mat;

	light_t				*lgt=NULL;
	vec_t				L,N;
	vec_t				V,R,H;
	rgb_t<double>			I_d, I_s;
	double				r,ndotl=0.0,n=32.0;
	list_t<light_t* >::iterator	litr;

  // get closest object, if any
  if(!(obj = model.find_closest(pos,dir,last_hit,dist)) || raydist > MAX_DIST)
    return;

#ifdef DEBUG
    std::cerr << "raydist: " << raydist << std::endl;
#endif

  // if hit, print info
  if(dist > 0) {
#ifdef DEBUG
    std::cerr << "Hit: " << obj->getname();

    std::cerr << "Dist = " << dist;
    std::cerr << "Loc = " << obj->getlast_hit() << std::endl;
#endif

    // accumulate distance travelled by ray
    raydist += dist;

    // get object material properties
    if((mat = model.getmaterial(obj->getmaterial())) != NULL) {
      ambient = mat->getamb();
      diffuse = mat->getdiff();
      specular = mat->getspec();
    }

    // ambient color
    color += 1.0/raydist * ambient;		// ambient scaled by ray dist

    N = obj->getnormal().norm();		// surf normal

    // view direction (direction from hit point to camera)
    //V = model.getviewpoint() - obj->getlast_hit();	// dir to camera
    //V = V.norm();
    V = -dir;						// dir to camera

    // diffuse component from each light...
    if(!diffuse.iszero()) {
      for(litr = model.lgts.begin(); litr != model.lgts.end(); litr++) {
        // pointer to light
        lgt = (light_t *)*litr;
  
        // light direction and distance
        L = lgt->getlocation() - obj->getlast_hit();
        r = L.len();				// distance to light
        L = L.norm();				// dir to light
  
        // angle with light
        ndotl = N.dot(L);
  
        // check visibility wrt light
        if(0.0 < ndotl && ndotl < 1.0) {
          // light color scaled by N . L
          I_d = 1.0/r * ndotl * lgt->getcolor();
  
          // add in diffuse contribution from light scaled by material property
          color += (I_d * diffuse);
        }
      }
  
      // clamp resultant color
      color.clamp(0.0,1.0);
    }

    // specular reflections (including specular highlights from lights)
    if(!specular.iszero()) {

	vec_t		hitpos(obj->getlast_hit());
	vec_t		out;
	rgb_t<double>	refcolor;

      // reflected ray direction
    //out = dir.reflect(N).norm();
      out = (((2.0 * N.dot(-dir)) * N) - -dir).norm();

      ray_trace(model,hitpos,out,refcolor,raydist,obj);

      // composite surface color: blending baesd on surface properties
      // (note that diffuse + specular should add up to 1)
      color  = (diffuse * color) + (specular * refcolor);

      // add in specular highlight from each light...
      for(litr = model.lgts.begin(); litr != model.lgts.end(); litr++) {
        // pointer to light
        lgt = (light_t *)*litr;

        // light direction and distance
        L = lgt->getlocation() - obj->getlast_hit();
        r = L.len();				// distance to light
        L = L.norm();				// dir to light

        // angle with light
        ndotl = N.dot(L);

        // check visibility wrt light
        if(0.0 < ndotl && ndotl < 1.0) {
          // specular reflection direction (and bisector)
        //R = (L.reflect(N)).norm();
          R = (((2.0 * ndotl) * N) - L).norm();
          H = (0.5 * (L + V)).norm();

          // light color scaled by (R . V)^n
          I_s = 1.0/r * pow(R.dot(V),n) * lgt->getcolor();
  
          // add in specular contribution from light scaled by material property
          color += (I_s * specular);
        }

        // clamp resultant color
        color.clamp(0.0,1.0);
      }
    }
  }
}
